/***************************************************************************
                          ADM_ffmp43.cpp  -  description
                             -------------------
                             
	Decoder for Divx3/4/..., using ffmpeg                             
                             
    begin                : Wed Sep 25 2002
    copyright            : (C) 2002 by mean
    email                : fixounet@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <math.h>
#include "config.h"
#ifdef USE_FFMPEG

#include "ADM_lavcodec.h"
#include "ADM_library/default.h"

#include "ADM_colorspace/colorspace.h"

#include "ADM_codecs/ADM_codec.h"

#include "ADM_codecs/ADM_ffmp43.h"
#include "ADM_gui/GUI_MPP.h"
 
#include "ADM_toolkit/toolkit.hxx"
uint32_t PP_DefaultValue=3;
uint32_t PP_DefaultStrength=5; 

void setPostProc(int v,int s)
{
	PP_DefaultValue=v;
	PP_DefaultStrength=s;
	printf("\n New default PP value  : %ld %ld",PP_DefaultValue,PP_DefaultStrength);

}   
//________________________________________________
void decoderFF::setParam( void )
{ 	 
		if( getMPParams( (int *)&_postProc,(int *) &_postProcStrength))
		{
				printf("\n new PP : %ld %ld \n",_postProc,_postProcStrength);
				updatePostProc();
			
		}

        return; // no param for ffmpeg
}
//________________________________________________
void					decoderFF::updatePostProc(void )		    
{
char stringMode[30];
			stringMode[0]=0;
			
	   if(_ppMode) {pp_free_mode(_ppMode);_ppMode=NULL;}

			
			if(_postProc&1) strcat(stringMode,"hb:a,");
			if(_postProc&2) strcat(stringMode,"vb:a,");
			if(_postProc&4) strcat(stringMode,"dr:a,");			
			
			if(strlen(stringMode)) 
				{
						stringMode[strlen(stringMode-1)]=0;
						_ppMode=pp_get_mode_by_name_and_quality(stringMode, _postProcStrength);;
						assert(_ppMode);

				}	   
				else    // if nothing is selected we may as well set back every thing to 0
				{
						_postProcStrength=0;
				}
}
//-------------------------------
decoderFF::decoderFF(uint32_t w,uint32_t h) :decoders(w,h)
{
	
//				memset(&_context,0,sizeof(_context));
				_context=NULL;
#if LIBAVCODEC_BUILD >= 4624
   	_context=avcodec_alloc_context();
#else
		_context=new AVCodecContext;
    memset(_context, 0, sizeof(AVCodecContext));
#endif	
				assert(_context);
				memset(&_frame,0,sizeof(_frame));
				
				_context->max_b_frames=0;
				
		  		_context->width=_w;
				_context->height=_h;							
			  	_context->pix_fmt				=PIX_FMT_YUV420P;
				//_context->debug=1;
			    
			    _internalBuffer=new uint8_t[w*h*3];
			    _postProc=PP_DefaultValue;
			    _postProcStrength=PP_DefaultStrength;
			    _ppMode=NULL;
			    _ppContext=NULL;
	   
	   		_ppContext=pp_get_context(_w, _h, 
#ifdef USE_MMX 
          PP_CPU_CAPS_MMX
#else
         0
#endif			    
			   );
			    assert(_ppContext);
			    updatePostProc();
			    printf("FFMpeg build : %d\n", LIBAVCODEC_BUILD);
}

//_____________________________________________________

decoderFF::~decoderFF()
{
		avcodec_close(_context);
		free(_context);
		delete [] _internalBuffer;
		 if(_ppMode) {pp_free_mode(_ppMode);_ppMode=NULL;}
		 if(_ppContext) {pp_free_context(_ppContext);_ppContext=NULL;}

		printf("FF base destroyed\n");
}
 
//-------------------------------

uint8_t     decoderFF::uncompress(uint8_t *in,uint8_t *out,uint32_t len,uint32_t *flagz)
										
{
 int got_picture=0;
 uint8_t *oBuff[3];
 int strideTab[3];
 int strideTab2[3];
 
		if(len==0) // Null frame, silently skip
			{
            	if(flagz) *flagz=0;
              printf("\n ff4: null frame\n");
							return 1;
			}
			     	
		 if(0> avcodec_decode_video(_context,
			  			&_frame,
             	&got_picture, in, len))
             {
					printf("\n error in FFMP43/mpeg4!\n");
					return 0;			 
				}	
		if(!got_picture)
		{
					printf("\n error in FFMP43/mpeg4!: got picture \n");
					GUI_Alert("Please retry with misc->Turbo off");
					return 0;			 			
		}		  
		
		// convert ffmpeg to our format : yv12
		uint8_t **src;
		if(_context->pix_fmt==PIX_FMT_YUV422P) 
		{
			uint32_t stride;
			uint8_t *xsrc;
			uint8_t *dst_y,*dst_u,*dst_v;
			printf("YUY2\n");
			printf(" Y: %d\n",_frame.linesize[0 ]);
			printf(" U: %d\n",_frame.linesize[1 ]);
			printf(" V: %d\n",_frame.linesize[2 ]);
			stride=  _frame.linesize[0 ];
			
			// we use post proc buffer for YUY2->YV12 conversion	
			// 
			oBuff[0]=_internalBuffer;
		 	oBuff[1]=_internalBuffer+_w*_h;
 		 	oBuff[2]=oBuff[1]+((_w*_h)>>2);		 
       			strideTab[0]=strideTab2[0]=_w;
			strideTab[1]=strideTab2[1]=_w>>1;
			strideTab[2]=strideTab2[2]=_w>>1;
			
			xsrc= _frame.data[0];	
			dst_y=oBuff[0];
			dst_u=oBuff[1];
			dst_v=oBuff[2];
			// y
            		for(uint32_t y=_h;y>0;y--)
			{				
				memcpy(dst_y,xsrc,_w);
				dst_y+=_w;
				xsrc+=_frame.linesize[0];
			}
			// u 
			xsrc= _frame.data[1];	
			for(uint32_t y=_h>>1;y>0;y--)
			{
				memcpy(dst_u,xsrc,_w>>1);
				dst_u+=_w>>1;
				xsrc+=_frame.linesize[1]*2;
				
			}
			// v
			xsrc= _frame.data[2];	
			for(uint32_t y=_h>>1;y>0;y--)
			{
				memcpy(dst_v,xsrc,_w>>1);
				dst_v+=_w>>1;
				xsrc+=_frame.linesize[2]*2;
			}
			  	src= (uint8_t **)oBuff;
				_frame.linesize[0 ]=_w;
				_frame.linesize[1 ]=_w>>1;
				_frame.linesize[2 ]=_w>>1;
			
		}
		else
		{
			if(!_postProc || !_postProcStrength)
			{
				src= (uint8_t **) _frame.data;			
			}
			else
			{ 	// we do postproc !
				// keep 	
		
		 				oBuff[0]=_internalBuffer;
		 				oBuff[1]=_internalBuffer+_frame.linesize[0]*_h;
 		 				oBuff[2]=oBuff[1]+((_frame.linesize[1]*_h)>>1);		 
        			for(uint32_t s=0;s<3;s++)
	          		{
			            strideTab[s]=strideTab2[s]=_frame.linesize[s];
            			}
		 		   pp_postprocess(
		      			_frame.data,
		        		strideTab,
		          		oBuff,   
		         		strideTab2,
		      			_w,
		        		_h,
		          		_frame.qscale_table,
		          		_frame.qstride,
		         		_ppMode,
		          		_ppContext,
		          		_frame.pict_type);
			  	src= (uint8_t **)oBuff;
			}
		}
		uint8_t *tmp;
		tmp=src[0];
		uint8_t *otmp= out;
		uint32_t stride;

		stride=  _frame.linesize[0 ];
				
		for(uint32_t y=_h;y>0;y--)
			{
				 	memcpy(otmp,tmp,_w);
				  	tmp+=stride;
   				  	otmp+=_w;
				}
		stride=  _frame.linesize[1 ];
		tmp= src[1];
		otmp= out+_w*_h+((_w*_h)>>2);
		
		for(uint32_t y=_h>>1;y>0;y--)
			{
				 	memcpy(otmp,tmp,_w>>1);
				  	tmp+=stride;
   				  	otmp+=_w>>1;
				}
		tmp= src[2];
		otmp= out+_w*_h;
		
		for(uint32_t y=_h>>1;y>0;y--)
			{
				 	memcpy(otmp,tmp,_w>>1);
				  	tmp+=stride;
   				  	otmp+=_w>>1;
				}		
				
		_lastQ=0; //_context->quality;		
		if(flagz)
		{
		//if(_context->key_frame)
		if(_frame.key_frame)
			*flagz=AVI_KEY_FRAME;
		else
			*flagz=0;
		}
				
 		return 1;
}



decoderFFDiv3::decoderFFDiv3(uint32_t w,uint32_t h)    :decoderFF(w,h) 
{
		      if (avcodec_open(_context, &msmpeg4v3_decoder) < 0) 
		      {
					printf(" Decoder init: FFMpeg MS MPEG43 video decoder failed!\n");
				}
				else
				{
					printf(" Decoder init: FFMpeg MS MPEG43 video decoder initialized!\n");
				}
}              

decoderFFMpeg4::decoderFFMpeg4(uint32_t w,uint32_t h)       :decoderFF(w,h) 
{
// force low delay as avidemux don't handle B-frames
		  
		  _context->flags|=CODEC_FLAG_LOW_DELAY;		
      if (avcodec_open(_context,& mpeg4_decoder) < 0) 
	      {
					printf(" Decoder init: FFMpeg MPEG4 video decoder failed!\n");
				}
				else
				{
					printf(" Decoder init: FFMpeg MPEG4 video decoder initialized!\n");
				}
}    
decoderFFDV:: decoderFFDV(uint32_t w,uint32_t h)       :decoderFF(w,h) 
{
      if (avcodec_open(_context,& dvvideo_decoder) < 0) 
	      {
					printf(" Decoder init: FFMpeg DV video decoder failed!\n");
				}
				else
				{
					printf(" Decoder init: FFMpeg DV video decoder initialized!\n");
				}
	
}
decoderFFMP42::decoderFFMP42(uint32_t w,uint32_t h)       :decoderFF(w,h) 
{
      if (avcodec_open(_context,& msmpeg4v2_decoder) < 0) 
	      {
					printf(" Decoder init: FFMpeg MP42 video decoder failed!\n");
				}
				else
				{
					printf(" Decoder init: FFMpeg MP42 video decoder initialized!\n");
				}
	
}
decoderFFH263::decoderFFH263(uint32_t w,uint32_t h)       :decoderFF(w,h) 
{
      if (avcodec_open(_context,& h263_decoder) < 0) 
	      {
					printf(" Decoder init: FFMpeg h263 video decoder failed!\n");
				}
				else
				{
					printf(" Decoder init: FFMpeg h263 video decoder initialized!\n");
				}
	
}

decoderFFhuff::decoderFFhuff(uint32_t w,uint32_t h,uint32_t l,uint8_t *d)       :decoderFF(w,h) 
{
      _context->extradata=(void *)d;
      _context->extradata_size=(int)l;
      
      if (avcodec_open(_context,& huffyuv_decoder) < 0) 
	      {
					printf(" Decoder init: FFMpeg huffy_uv video decoder failed!\n");
				}
				else
				{
					printf(" Decoder init: FFMpeg huffy_uv video decoder initialized!\n");
				}
	
}
decoderFFWMV2::decoderFFWMV2(uint32_t w,uint32_t h,uint32_t l,uint8_t *d)       :decoderFF(w,h) 
{
      _context->extradata=(void *)d;
      _context->extradata_size=(int)l;
      
      if (avcodec_open(_context,& wmv2_decoder) < 0) 
	      			{
					printf(" Decoder init: FFMpeg WMV2 video decoder failed!\n");
				}
				else
				{
					printf(" Decoder init: FFMpeg WMV2 video decoder initialized!\n");		
				}
	
}
decoderFFcyuv::decoderFFcyuv(uint32_t w,uint32_t h,uint32_t l,uint8_t *d)       :decoderFF(w,h) 
{
      _context->extradata=(void *)d;
      _context->extradata_size=(int)l;
      
      if (avcodec_open(_context,& cyuv_decoder) < 0) 
	      			{
					printf(" Decoder init: FFMpeg cyuv video decoder failed!\n");
				}
				else
				{
					printf(" Decoder init: FFMpeg cyuv video decoder initialized!\n");		
				}
	
}


#endif
