/***************************************************************************
                          ADM_edit.cpp  -  description
                             -------------------
    begin                : Thu Feb 28 2002
    copyright            : (C) 2002 by mean
    email                : fixounet@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "fourcc.h"
#include "avio.hxx"
#include "config.h"
#include "ADM_editor/ADM_edit.hxx"
#include "ADM_inpics/ADM_pics.h"
#include "ADM_nuv/ADM_nuv.h"
#include "ADM_h263/ADM_h263.h"
#include "ADM_mpeg2dec/ADM_mpegscan.h"
#include "ADM_toolkit/toolkit.hxx"
#include "ADM_editor/ADM_edit.hxx"
#include "ADM_video/ADM_genvideo.hxx"
#include "ADM_filter/video_filters.h"
#include "assert.h"


int  GUI_mpegIndexer(char **mpegFile,char **indexFile,int *aid,int already=0);
void GUI_mpegIndexerPrefill(char *name);
extern uint8_t indexMpeg(char *mpeg,char *file,uint8_t aid);

//
//

#define TEST_MPEG2DEC

ADM_Composer::ADM_Composer(void)
{
    	_nb_segment = 0;
    	_total_frames = 0;
     	_audioseg=0;
    	_audiooffset=0;
        _lastseg=99;
        _lastframe=99;
        _cached=0;

}
// it is only used for some codec that needs the extra data to be correctly
// initialized.
uint8_t ADM_Composer::getExtraHeaderData(uint32_t *len, uint8_t **data)
{
aviHeader *hack;
	//we assume there is only one file
	hack=(aviHeader *)_segments[0]._aviheader;
	return hack->getExtraHeaderData(len,data);

}
uint8_t	ADM_Composer::getAudioExtra(uint32_t *l,uint8_t **d)
{
	*l=0;
	*d=NULL;
	if(!_nb_segment) return 0;
	if(! _segments[0]._audiostream) return 0;
	return _segments[0]._audiostream->extraData(l,d);


}
//
//	Delete all uniq segment
//
void ADM_Composer::deleteAllSegments()
{
	if(!_nb_segment) return;
	// first removes duplicates
 	for(uint32_t seg=0;seg<(_nb_segment-1);seg++)
	{
		for(uint32_t segb=seg+1;segb<_nb_segment;segb++)
		{
			if(  _segments[seg]._aviheader==_segments[segb]._aviheader)
   				{
                   _segments[segb]._nb_video_frames=0;
         		}
   		}
   	}

	crunch();

 	// now we have only uniq avi segments
	for(uint32_t seg=0;seg<_nb_segment;seg++)
	{
		
    		
      		if( _segments[seg]._audiostream)
        	{
           	if( _segments[seg]._audiostream->isDestroyable())
		    			delete _segments[seg]._audiostream;
       	}
       	if( _segments[seg].decoder)
        		delete _segments[seg].decoder;	
          // prevent from crashing
           _segments[seg]._aviheader->close();
    		delete _segments[seg]._aviheader;
	}

 	memset(   _segments,0,sizeof(_segments));
	
}
ADM_Composer::~ADM_Composer()
{
	   deleteAllSegments();

}

/*
   			Return Magic : 4*4 bytes first

*/
uint8_t ADM_Composer::getMagic(char *name,uint32_t *magic)
{
 	       FILE *tmp;
			uint8_t ret=1;

			tmp=fopen(name,"rb");
			if(!tmp) return 0;

			if(4!=fread(magic,4,4,tmp))
							ret=0;

			fclose(tmp);
			return ret;

}

/*
   		Tries to identify the incoming stream using its 4 magic word


*/

uint8_t ADM_Composer::identify(char *name, fileType *type)
{
uint32_t magic[4];
       *type=Unknown_FileType;
		if(!getMagic(name,magic)) return 0;

		// now try to identifies the filetype by its magic
		if( fourCC::check(magic[0],(uint8_t *)"RIFF"))
			{
            		printf(" \n Riff file detected...");
					if( fourCC::check(magic[2],(uint8_t *)"AVI "))
						{
				            		printf(" \n AVI file detected...\n");
					              *type=AVI_FileType;
									return 1;
                      }
	            		printf(" \n Unknown Riff...");
						return 0;
			}
			
		if( fourCC::check(magic[0],(uint8_t *)"IDXM"))
					{
                      	printf(" \n DVD2AVI Mpeg file detected...\n");
					              *type=MpegIdx_FileType;
									return 1;
					}
			
		if( fourCC::check(magic[0],(uint8_t *)"TOC "))
					{
                      	printf(" \n Mpeg file detected...\n");
					              *type=Mpeg_FileType;
									return 1;
					}

			if(                // fixme: put a mask
				(magic[0]==0xba010000) || 
				  (magic[0]==0xb3010000)
				)
					{
                      	printf(" \n Mpeg file detected...\n");
					              *type=Mpeg_FileType;
									return 1;
					}
         if( fourCC::check(magic[0],(uint8_t *)"Nupp")
	 	|| fourCC::check(magic[0],(uint8_t *)"Myth")
	 )
			{
                      	printf(" \n Nuppelvideo file detected...\n");
					              *type=Nuppel_FileType;
									return 1;			
			}
			if( magic[0]==0x05364d42)
							{

                      			printf(" \n BMP file detected...\n");
					              *type=BMP_FileType;
									return 1;
						}

		if( magic[0]==0x02800000)
		{
                 	printf(" \n Raw H263 file detected...\n");
			*type=H263_FileType;
			return 1;
		}
		if( magic[0]==0x00010000)
		{
                 	printf(" \n Raw mpeg4 file detected...\n");
			*type=Mp4_FileType;
			return 1;
		}
		printf("\n unrecognized file detected...\n");
		return 0;
}
//______________________________________________
// Save the config, including name, segment etc...
//______________________________________________

uint8_t ADM_Composer::saveWorbench(char *name)
{
char *tmp;

	 if(!_nb_segment) return 1;
	 
	 FILE *fd;
	 
	 fd=fopen(name,"wt");
		 
	 fprintf(fd,"Avidemux Config file 0 : %02ld segments\n",_nb_segment);
	 
	 for(uint32_t i=0;i<_nb_segment;i++)
	 {
		   fprintf(fd,"Name : %s\n", _segments[i]._aviheader->getMyName());
		   fprintf(fd,"Start: %lu\n", _segments[i]._start_frame);
		   fprintf(fd,"Size : %lu\n", _segments[i]._nb_video_frames);		 
		}
	  fclose(fd);
	  // try to load filters
	tmp = (char*)malloc(strlen(name)+5);
	assert(tmp);
	strcpy(tmp,name);
	strcat(tmp,".flt");
	filterSave(tmp,1);
        free(tmp);
	return 1;		
}
//______________________________________________
// Save the config, including name, segment etc...
//______________________________________________

uint8_t ADM_Composer::loadWorbench(char *name)
{
char str[400];
char str_tmp[400];
uint32_t nb,start,size,j;
int match;
char a;
FILE *fd;
char *tmp;

	 cleanup();
	 
	 fd=fopen(name,"rt");
		 
	 fscanf(fd,"Avidemux Config file 0 : %lu segments\n",&nb);
	 printf("WKB Reading workbench, %ld segments to go\n",nb);
	 
	 for(uint32_t i=0;i<nb;i++)
	 {
		 
		  // fscanf(fd,"Name : %s\n",str);  // Fix for dir/name with space in it
		   fgets(str_tmp,390,fd);
		   a=str_tmp[strlen(str_tmp)-1];
		   if((a==0x0a) || (a==0x0d)) str_tmp[strlen(str_tmp)-1]=0;
		   strcpy(str,str_tmp+7),
			 fscanf(fd,"Start: %lu\n", &start);
		   fscanf(fd,"Size : %lu\n", &size);			   
		   // search if we got it already
		   match=-1;
		   printf("WKB Segment  : %02ld start : %ld, size :%ld, file :*%s*\n",i,
		   				start,size,str);
		   for( j=0;j<_nb_segment;j++)
		   {
				 	if(_segments[j]._aviheader->getMyName()==NULL)
				  {
							printf("\nWKB***************** Warning ! Null name!\n");
					}
					else
					if(!strcmp( str, _segments[j]._aviheader->getMyName()))
					{
						match=j;
					} 
			 }
			 if(match==-1)
			 {
				  if(!addFile(str))
				  {
							GUI_Alert("Error loading one the segments!");
							printf("Error loading one the segments!:%s\n",str);

							goto abt;
					}
					_segments[_nb_segment-1]._start_frame=start;
					_segments[_nb_segment-1]._nb_video_frames=size;					
				}
				else
				{
					printf("WKB->already loaded as seg %ld\n",j);
					memcpy(&_segments[_nb_segment],&_segments[match],
								sizeof(_segments[_nb_segment]));
					
					_segments[_nb_segment]._start_frame=start;
					_segments[_nb_segment]._nb_video_frames=size;
					_nb_segment++;
				}
				
			
				
				
				updateAudioTrack(_nb_segment-1);				
		}
abt:								
	  fclose(fd);
	  _total_frames=0;
    // update total frames
    for(uint32_t n=0;n<_nb_segment;n++)
    {
			_total_frames+= _segments[n]._nb_video_frames;
		}
		dumpSeg()	  ;
		printf("\n Total frames : %lu \n",_total_frames);
		
// try to load filters
	tmp = (char *)malloc(strlen(name)+5);
	assert(tmp);
	strcpy(tmp,name);
	strcat(tmp,".flt");
	filterLoad(tmp,1);
	free(tmp);
	return 1;		
}
//______________________________________________
// load (append) a file
//______________________________________________
uint8_t ADM_Composer::addFile(char *name, uint8_t mode)
{
    uint8_t ret = 0;
    aviInfo info;
    WAVHeader *_wavinfo;
    aviHeader  *tmp;
	fileType type=Unknown_FileType;



    assert(_nb_segment < MAX_SEG);

		if(!identify(name,&type)) return 0;

		switch(type)
			{
					case Mp4_FileType:
						 _segments[_nb_segment]._aviheader = new mp4Header;
						ret = _segments[_nb_segment]._aviheader->open(name);
						break;

					case H263_FileType:
						 _segments[_nb_segment]._aviheader = new h263Header;
						ret = _segments[_nb_segment]._aviheader->open(name);
						break;
					case AVI_FileType:
					    	_segments[_nb_segment]._aviheader = tmp=new aviHeader;
				       		if(mode==1)
			             			ret = tmp->openBrokenAVI(name);
			           		else
							ret = _segments[_nb_segment]._aviheader->open(name);
						break;

					case Nuppel_FileType:
								 _segments[_nb_segment]._aviheader = new nuvHeader;
								ret = _segments[_nb_segment]._aviheader->open(name);
								break;

	              case BMP_FileType:
					        _segments[_nb_segment]._aviheader = new picHeader;
								ret = _segments[_nb_segment]._aviheader->open(name);
								break;

					case Mpeg_FileType:
							   		// then try mpeg
						           if(GUI_Question("This looks like mpeg\n Do you want to index it?"))
						           {
												 		char *idx,*mname;
												    int track;
												    
												   		GUI_mpegIndexerPrefill(name);
												   		if(GUI_mpegIndexer(&mname,&idx,&track,1))
												   		{
														  	if((mname==NULL ) || (idx==NULL))
															 	{	
																	 	GUI_Alert("Select files!");
																	  
				 														return 0 ;
																}
			 													printf("\n indexing :%s to \n%s\n",mname,idx);
   												    	GUI_StopWorking(); // ugly

																if(indexMpeg(mname,idx,(uint8_t ) track))
																	GUI_Alert("Reopen the index");	 			 
																GUI_Working();														 
															}												 
												}
					       		return 0;
									break;
					case MpegIdx_FileType:
									_segments[_nb_segment]._aviheader = new mpeg2decHeader;
									ret = _segments[_nb_segment]._aviheader->open(name);	
									break;

					default:
								if(   type== Unknown_FileType)
										{
                                    		printf("\n not identified ...\n");
										}
									else
                                    		printf("\n successfully identified but no loader support detected...\n");
                       return 0;
         }
		// check opening was successful
			if (ret == 0)
					      {
						        printf("\n Attempt to open %s failed!\n",name);
							  	delete _segments[_nb_segment]._aviheader;;
								return 0;
							}

    // else update info
    _segments[_nb_segment]._aviheader->getVideoInfo(&info);
    _segments[_nb_segment]._aviheader->setMyName(name);
//    fourCC::print( info.fcc );
    _total_frames += info.nb_frames;
    _segments[_nb_segment]._nb_video_frames = info.nb_frames;
    _segments[_nb_segment]._orig_nb_video_frames = info.nb_frames;

    _segments[_nb_segment]._start_frame = 0;

    // and update audio info
    //_________________________
    _wavinfo = _segments[_nb_segment]._aviheader->getAudioInfo();	//wavinfo); // will be null if no audio
    if (!_wavinfo)
      {
	  printf("\n *** NO AUDIO ***\n");
	  _segments[_nb_segment]._audiostream = NULL;
    } else
      {
	   _segments[_nb_segment]._aviheader->getAudioStream(&_segments[_nb_segment]._audiostream);
	   _segments[_nb_segment]. _audio_size=_segments[_nb_segment]._audiostream->getLength() ;
       _segments[_nb_segment]._audio_start=0;
      }

      printf("\n Decoder FCC: ");
      fourCC::print(info.fcc);
      // ugly hack
      if(info.fps1000>2000*1000)
      {
      	printf(" FPS too high, switching to 25 fps hardcoded\n");
	info.fps1000=25*1000;
	updateVideoInfo(&info);
      }
      uint32_t l;
      uint8_t *d;
      _segments[_nb_segment]._aviheader->getExtraHeaderData(&l,&d);
      _segments[_nb_segment].decoder=getDecoder(
       								info.fcc,
       								info.width,
      								info.height,
								l,d);

    _nb_segment++;
    return 1;
}

//
// Set decoder options (post processing etc etc etc)
//
uint8_t ADM_Composer::setDecodeParam( void )
{
	if( _nb_segment)
 	   _segments[0].decoder->setParam();
    return 1;

}
uint8_t ADM_Composer::cleanup(void)
{
   deleteAllSegments();



    _nb_segment = 0;
    _total_frames = 0;
    return 1;
}

//______________________________________________________
//  Remove frames , the frame are given as seen by GUI
//  We remove from start to end -1
//
//______________________________________________________

uint8_t ADM_Composer::removeFrames(uint32_t start, uint32_t end)
{

    uint32_t seg1, seg2, rel1, rel2;
   uint8_t   lastone=0;
   
   if(end==_total_frames-1) lastone=1;
    
    // sanity check
    if (start > end)
	return 0;
    //if((1+start)==end) return 0;

    // convert frame to block, relative frame
    if (!convFrame2Seg(start, &seg1, &rel1) ||
	!convFrame2Seg(end, &seg2, &rel2))
      {
	  assert(0);
      }
    // if seg1 != seg2 we can just modify seg1 and seg2
    if (seg1 != seg2)
      {
	  // remove end of seg1

	  removeFrom(rel1, seg1, 1);
	  //  delete in between seg
	  for (uint32_t seg = seg1 + 1; seg < (seg2); seg++)
	      _segments[seg]._nb_video_frames = 0;
	  // remove beginning of seg2
	  removeTo(rel2, seg2, lastone);
    } else
      {
	  // it is in the same segment, split it...
	  // complete seg ?
	  if ((rel1 == _segments[seg1]._start_frame)
	      && (rel2 ==
		  (_segments[seg1]._start_frame +
		   _segments[seg1]._nb_video_frames - 1)))
	    {
		_segments[seg1]._nb_video_frames = 0;
	  } else
	    {
		// split in between.... duh !
		duplicateSegment(seg1);
		//
		removeFrom(rel1, seg1, 1);
		removeTo(rel2, seg1 + 1, lastone);
	    }
      }

    // Crunch
    crunch();
    sanityCheck();
    // Compute total nb of frame
    _total_frames = computeTotalFrames();
    printf("\n %lu frames ", _total_frames);
    return 1;

}

//____________________________________
//      Duplicate a segment
//____________________________________

uint8_t ADM_Composer::duplicateSegment(uint32_t segno)
{

    for (uint32_t i = _nb_segment; i > segno; i--)
      {

	  memcpy(&_segments[i], &_segments[i - 1], sizeof(_SEGMENT));

      }
    _nb_segment++;
    return 1;


}

//____________________________________
//      Remove empty segments
//____________________________________
uint8_t ADM_Composer::crunch(void)
{
    uint32_t seg=0;
    while (seg < _nb_segment)
      {
	  if (_segments[seg]._nb_video_frames == 0)
	    {
       //

		for (uint32_t c = seg + 1; c < _nb_segment; c++)
		  {
		      memcpy(&_segments[c - 1], &_segments[c],
			     sizeof(_SEGMENT));
		  }
		_nb_segment--;

	  } else
	    {
		seg++;
	    }
      }
      // Remove last seg if there is only one frame in it
      if(_nb_segment)
      {
      if( _segments[_nb_segment-1]._nb_video_frames == 1)
      	{
            _nb_segment--;
         }
       }
    return 1;

}

//____________________________________
//      Remove empty segments
//____________________________________
uint32_t ADM_Composer::computeTotalFrames(void)
{
    uint32_t seg, tf = 0;
    for (seg = 0; seg < _nb_segment; seg++)
      {
	  tf += _segments[seg]._nb_video_frames;

      }

    return tf;
}

//____________________________________
//      Remove empty segments
//____________________________________
void ADM_Composer::dumpSeg(void)
{
    uint32_t seg;
    printf("\n______________________");
    for (seg = 0; seg < _nb_segment; seg++)
      {
	  printf("\n Seg : %lu, start :%lu, size:%lu audio size : %lu audio start : %lu",
		 seg,
		 _segments[seg]._start_frame,
		 _segments[seg]._nb_video_frames,
        _segments[seg]. _audio_size,
        _segments[seg]. _audio_start);

      }


}

// Clear from position to/from
//
// 0------To------end
//  xxxxxx removed

uint8_t ADM_Composer::removeTo(uint32_t to, uint32_t seg, uint8_t included)
{
    uint32_t ref;

    assert(checkInSeg(seg, to));
    ref = _segments[seg]._start_frame;
    _segments[seg]._start_frame = to;
    if (included)
			_segments[seg]._start_frame++;
    _segments[seg]._nb_video_frames -= (_segments[seg]._start_frame - ref);


    updateAudioTrack(seg);

//---------------------------------------



    return 1;
}
//
//	Update the real size of audio track by computing the
// delta between sync @end and sync@begin
//

uint8_t ADM_Composer::updateAudioTrack( uint32_t seg)
{
     // audio sync
    uint32_t pos_start,pos_end,off, tf;
              // Mika
    if(! _segments[seg]._audiostream) return 1;
    // If we cannot go to sync point start --> no need to continue
    // It can happen if audio track is shorter than video
    if(!audioGoToFn(  seg,   _segments[seg]._start_frame,&off))
    {
       	_segments[seg]._audio_size=0;
        	printf(" cannot seek audio tp frame : %lu\n",seg);
         	return 1;

      }
    pos_start= _segments[seg]._audiostream->getPos();

    // Now try to go to the end...
    // if it fails, try previous frame stamp
    tf= _segments[seg]._start_frame+_segments[seg]._nb_video_frames;//-1;

    while((! audioGoToFn(  seg, tf,&off)
    	&& ( (tf-1)>_segments[seg]._start_frame)))
     {
         printf(" trying to sync on frame %lu\n",tf);
	     tf--;
      }
    								
    pos_end= _segments[seg]._audiostream->getPos();

     _segments[seg]._audio_size=pos_end-pos_start;
     _segments[seg]._audio_start=pos_start;

     printf(" Audio start : %lu end : %lu size : %lu\n",pos_start,pos_end,_segments[seg]._audio_size);

     return 1;

#warning FIXME, does not work if audio track is shorter

}

//
// 0------From------end
//            xxxxxx removed

uint8_t ADM_Composer::removeFrom(uint32_t from, uint32_t seg,
				 uint8_t included)
{
    assert(checkInSeg(seg, from));
    _segments[seg]._nb_video_frames = (from - _segments[seg]._start_frame);

    if (!included)
	_segments[seg]._nb_video_frames++;

   updateAudioTrack(seg);
    return 1;
}

//__________________________________________________
// check that the given frame is inside the segment
//__________________________________________________
uint8_t ADM_Composer::checkInSeg(uint32_t seg, uint32_t frame)
{
    if (frame < _segments[seg]._start_frame)
	return 0;
    if (frame >
	(_segments[seg]._nb_video_frames + _segments[seg]._start_frame))
	return 0;
    return 1;

}

uint8_t ADM_Composer::sanityCheck(void)
{
    uint32_t ref, seg;

    for (seg = 0; seg < _nb_segment; seg++)
      {
	  ref =
	      _segments[seg]._start_frame +
	      _segments[seg]._nb_video_frames - 1;
	  assert(ref <= _segments[seg]._orig_nb_video_frames);

      }
    return 1;


}
void ADM_Composer::propagateBuildMap( void )
{
	  if(_nb_segment)
	  {
			_segments[0]._audiostream->buildAudioTimeLine();
		}
	
}

//
//
