/***************************************************************************
                          ADM_mpegindexer.cpp  -  description
                             -------------------
		Mix between mpeg2dec code, vdub one and mine.         
		
		The audio stream is a bit complicated
		if C0 C1 ... mpeg audio directly
		if 0... 7 -> AC3 track no i, the conversion to sub stream ID / PRIVATE STREAM 1 will be done
						internally by the demuxer		                                        
                             
    begin                : Tue Oct 15 2002
    copyright            : (C) 2002 by mean
    email                : fixounet@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef __FreeBSD__
          #include <sys/types.h>
#endif
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "config.h"
#include <math.h>

#include "ADM_library/default.h"
#include "avifmt.h"
#include "avifmt2.h"

#include "ADM_editor/ADM_Video.h"
#include "ADM_audio/aviaudio.hxx"

#include "ADM_library/fourcc.h"
#include "ADM_mpegindexer/ADM_mpegparser.h"
#include "ADM_mpeg2dec/ADM_mpegpacket.h"
#include "ADM_mpeg2dec/ADM_mpegpacket_PS.h"

#include "ADM_toolkit/toolkit.hxx"
#include "ADM_toolkit/filesel.h"
void mpegToIndex2(char *name);
void mpegToIndex(char *name)  ;
//static void mpegPictureScan2(char *name);

//static void skip_block( void );
//static int skipPacketHeader( void );
//static void videoPacket(void);
//static void audioPacket(void);
//static void picture(void);
//static void packStart ( void );
//static void infoStream( void ) ;
//static void sysHeader ( void );
//static void privatePacket(void);
uint8_t indexMpeg(char *mpeg,char *file);

static const uint32_t FPS[16]={
		0,			// 0
		23976,		// 1 (23.976 fps) - FILM
		24000,		// 2 (24.000 fps)
		25000,		// 3 (25.000 fps) - PAL
		29970,		// 4 (29.970 fps) - NTSC
		30000,		// 5 (30.000 fps)
		50000,		// 6 (50.000 fps) - PAL noninterlaced
		59940,		// 7 (59.940 fps) - NTSC noninterlaced
		60000,		// 8 (60.000 fps)
		0,			// 9
		0,			// 10
		0,			// 11
		0,			// 12
		0,			// 13
		0,			// 14
		0			// 15
	};



//static char *outname;
//static char *inname;
mParser *parser;
uint32_t nb_pic=0,skip=0,nb_aud=0;;
uint32_t sz_pic=0,sz_aud=0;

uint64_t pos_start=0;
FILE *_out;
uint8_t bfe[720*576*2];

uint8_t indexMpeg(char *mpeg,char *file,uint8_t audioid)
{
	static const char Type[5]={'X','I','P','B','D'};
	FILE *out=NULL;
	uint32_t val;
	uint32_t hnt;;
//	uint32_t in;
	uint64_t pos;
	uint64_t abspos;

	uint32_t w,h,fps ;	
	uint32_t nb_iframe=0;
//	uint32_t nb_pframe=0;
//	uint32_t nb_bframe=0;
//	uint32_t first=0;
	uint32_t total_frame=0;
	uint32_t seq_found=0;
	uint64_t lastI=0;
	uint32_t update=0;
    uint8_t streamid,audiostreamid=0x00;
	mParser *parser;
	ADM_mpegDemuxer *demuxer;

	audiostreamid=audioid;
	parser=new mParser();	
	
	printf("audio Id: %x\n",audiostreamid);
	if(!parser->open(mpeg))	
		{
				GUI_Alert("Error reading mpeg !");
				delete parser;
				return 0;
		}
		uint8_t token;
	
	parser->sync(&token);
	
	// first identify stream type (ES/PS)
	delete parser;

	switch(token)
		{
				case 0xb3: // video es
									printf("Video Elementary stream");
									demuxer=new ADM_mpegDemuxerElementaryStream;
									break;
				case 0xba : // program stream ?	(vob/mpeg/VCD)
									printf("Program stream");

									demuxer=new ADM_mpegDemuxerProgramStream(0xe0,audiostreamid);
									break;				 	
				default:
								printf("\n unrecognized stream\n");
								return 0;
												
		}
	// ok now we have it.
	demuxer->open(mpeg);		
	out=fopen(file,"wt");
	if(!out)
	{
			printf("\n Error : cannot open index !");
			delete demuxer;
			return 0;
	}
	fprintf(out,"IDXM ");
	if(token==0xb3) fprintf(out,"E XX\n");
	else   fprintf(out,"P XX\n");
	fprintf(out,"000000000000\n");	
	fprintf(out,"1\n");
	fprintf(out,"%s\n",mpeg);    
	GUI_Working();
	while(1)
	{
				val=0;
				hnt=0;
				
				// sync on the video stream is done by demuxer, now get on the infos
				//
				if(!demuxer->sync(&streamid)) break;
				
				demuxer->getpos(&pos);
				abspos=demuxer->getAbsPos();
				update++;
			//	printf("\n sync found : %d",streamid);
				if(update>100)
					{
						if(GUI_UpdateWorking(demuxer->getAbsPos()>>16,demuxer->getSize()>>16))
						{
							// abort;
							goto stop_found;	
							
						}
						update=0;
					}
				switch(streamid)
					{
						  case 0xB9: // sequence end						 
                                        printf("\n End seq\n");
                                        goto stop_found;
                                        break;
						  case 0xB3: // sequence start						 
											  if(seq_found)
											  	{
													 	demuxer->forward(8);
													  break;
													}
											  seq_found=1;
								             val=demuxer->read32i();
							   					w=val>>20;
							        			w=((w+15)&~15);
								        		h= (((val>>8) & 0xfff)+15)& ~15;
								      			fps= FPS[val & 0xf];
						    			        demuxer->forward(4);
						         			   printf("\n %ld x %ld at %ld fps\n",w,h,fps);
					           			 	fprintf(out,"%ld %ld %ld\n",w,h,fps);

						            break;
							case 0xb8: // GOP						            
											//	printf("GOP\n");
												demuxer->forward(4);
												break;
							case 0x00 : // picture
							                //  printf("PIC\n");
												uint8_t ftype;
												
												total_frame++;
												val=demuxer->read16i();
												ftype=7 & (val>>3);
												switch(ftype)
													{
														case 2:
														case 3:
														case 4:
																	if(total_frame==1)
																		{
																				total_frame=0;
																				continue;
																		}
														case 1:// printf(" Pic : I\n");
																		//	printf("\n%c",Type[ftype]);
																			if(total_frame!=1)
																				{
																	          fprintf(out," %llu\n",pos-lastI-4);
																	         
																				}
																				
																				
														          fprintf(out,"%c %010llX",Type[ftype],pos-4); 
														          if(ftype==1) // I frame
														          {
																			  fprintf(out," %010llX %10lx",abspos-4,demuxer->getOtherSize()); 	
																	}
														          lastI=pos-4;
														          nb_iframe++;
																			break;
															
													}
												
												break;	
							default:		
												break;																																																			
					}
		}	
stop_found:		
//	 	fprintf(out," %d\n\n",demuxer->getSize()-lastI-4);
	 	fprintf(out," 0 \n\n");
       printf("\n end of stream...\n");
	  // update # of frames
	    fseek(out,0,SEEK_SET);
	  	fprintf(out,"IDXM ");
		if(token==0xb3) fprintf(out,"E %02x\n",audiostreamid);
			else   fprintf(out,"P %02x\n",audiostreamid);
		fprintf(out,"%012lX\n",nb_iframe);	

		fclose(out);
		GUI_StopWorking();
	  printf("\n Total : %lu frames\n",total_frame);
	  delete demuxer;
	  return 0;
}


/*
void sysHeader ( void )
{
uint32_t size;//,part2;

		size=parser->read16i();
		parser->forward(size);

		printf("  sys header size : %lu\n",size);
		

}
*/

/*
void packStart ( void )
{
	parser->forward(8);
}
*/

/*
void infoStream( void )
{
		

}
*/

/*
   		Audio packet ...
		Raw stream, we can dump it whenever we want

*/
/*
void audioPacket(void)
{
int psize;
//unsigned long long pos;
		psize=skipPacketHeader();
#ifdef AV_VERBOSE		
		printf(" audio packet : %d\n",psize);
#endif		
		parser->forward(psize);
		
		sz_aud+=psize;
		nb_aud++;
}
*/

/*
void privatePacket(void)
{
uint32_t psize;
//uint32_t un;
//uint64_t pos;

		psize=skipPacketHeader();			
		printf(" Private packet : %ld\n",psize);

		parser->forward(psize);

}
*/

/*
void videoPacket(void)
{
int psize;
//uint32_t val,un;
//uint64_t end;
//uint32_t s;

		psize=skipPacketHeader();
		sz_pic+=psize;
#ifdef AV_VERBOSE		
		printf(" video packet : %d\n",psize);
#endif		
				   	nb_pic++;
    
		parser->forward(psize);
//		parser->read32(psize,bfe);
//		fwrite(bfe,psize,1,_out);
		
		return;
		

}
*/

//________________________________
// Remove packet header (PTS etc)
// Not needed for audio packet
//________________________________

/*
int skipPacketHeader( void )
{
uint32_t un ,deux, size;
uint8_t c;
			size=parser->read16i();
      size--;		
			// remove padding
			while((c=parser->read8i()) == 0xff) 
								{
									size--;
								}

			// now look at  STD buffer size if present
			// 01xxxxxxxxx
			if ((c>>6) == 1) 
			{	// 01
						size-=2;
						parser->read8i();			// skip one byte
						c=parser->read8i();   // then another
			}			
			// PTS/DTS
			switch(c>>4)
				{
				case 2:
						// 0010 xxxx
								size -= 4;
								parser->forward(4);
								break;
				case 3:
							// 0011 xxxx
									size -= 9;
									parser->forward(9);
									break;
				case 1:
							// 0001 xxx		
									{     // PTSDTS=01 not allowed
											assert(0);
									}
									break;
				}

//				size--; test resync
    	return size;
}
*/

//________________________________
// 	skip until 000001 is found (3 bytes)
// 	to quicky get ride of padding for example
//
//________________________________
/*
void skip_block( void )
{

}
*/


